import numpy as np
from lbbutils.metrics.mi import _mi
from PIL import Image


def _q_mi(im1: np.ndarray, im2: np.ndarray, fim: np.ndarray):
    """
        Ref: Comments on "Information measure for performance of image fusion"
        By M. Hossny et al.
        Electronics Letters Vol. 44, No.18, 2008
    :param im1:
    :param im2:
    :param fim:
    :return:
    """
    [I_fx, H_xf, H_x, H_f1] = _mi(im1, fim)
    [I_fy, H_yf, H_y, H_f2] = _mi(im2, fim)

    res = 2 * (I_fx / (H_f1 + H_x) + I_fy / (H_f2 + H_y))
    return res


if __name__ == '__main__':
    # img1 = torch.rand((1, 1, 256, 256))
    # img2 = torch.rand((1, 1, 256, 256))
    # lam = lambda x: x.data.squeeze(0).permute((1, 2, 0)).numpy()
    # x_ = lam(img1)
    # y_ = lam(img2)
    # ret1 = normalize(x_)
    # ret2 = normalize(y_)
    m1 = np.array(Image.open('../test/res/left.png'), dtype=np.float64)
    m2 = np.array(Image.open('../test/res/right.png'), dtype=np.float64)
    fim = np.array(Image.open('../test/res/fim.png'), dtype=np.float64)
    ret = _q_mi(m1, m2, fim)
    a = 1



import numpy as np
import torch
from torchvision import utils
import os
import time


class Visualization:
    def __init__(self, args):
        self.args = args
        from visdom import Visdom
        self.vis = Visdom()
        self.save_log = os.path.join(args.save_path, 'log', 'loss_log.txt')
        self.save_img = os.path.join(args.save_path, 'img')
        os.makedirs(os.path.dirname(self.save_log), exist_ok=True)
        os.makedirs(self.save_img, exist_ok=True)
        with open(self.save_log, 'w') as f:
            f.write('====================Train Loss(%s)========================\n' % time.strftime('%c'))

    def print_current_loss(self, e, iter, loss, iter_time, issave=True):
        """
        打印第e个epoch下第iter个batch损失值和的耗时
        :param issave:  是否保存loss记录。（默认保存）
        :param e:
        :param iter:
        :param loss:
        :param iter_time: 训练这个batch所需要的时间
        :return:
        """
        message = f'[Epoch {e}/{self.args.epochs}]-[batch {iter}/{self.args.len_dataloader}] ' \
                  f'\tLoss: {loss}\ttime: {iter_time}'
        print(message)
        if issave == True:
            with open(self.save_log, 'a') as f:
                f.write(message + '\n')

    def plot_current_loss(self, e, iter, loss):
        """
        将loss连成连线
        :param e:
        :param iter:
        :param loss:
        :return:
        """
        if not hasattr(self, 'plot_data'):
            self.plot_data = {'x': [], 'y': []}
        self.plot_data['x'].append(e + iter / self.args.len_dataloader)
        self.plot_data['y'].append(loss)
        x = np.stack(self.plot_data['x'])
        y = np.stack(self.plot_data['y'])
        self.vis.line(Y=y,
                      X=x,
                      opts={
                          'title': self.args.name + ' loss over time',
                          'xlabel': 'epoch',
                          'ylabel': 'loss'
                      })

    def display_current_result(self, e, visuals: dict, issave=True):
        """
        显示图像
        :param issave: boolean 是否保存网络输入输出、target和label
        :param epoch:
        :param visual:
        :return:
        """
        ncols = len(visuals)
        imgs = [np.array(img.cpu().detach()[0]) for img in visuals.values()]
        self.vis.images(imgs, nrow=ncols, padding=2, opts=dict(title=self.args.name + ' images'))
        if issave:
            for i, img in visuals.items():
                utils.save_image(img, os.path.join(self.save_img, f'{e}_{i}.png'))


if __name__ == '__main__':
    from lbbutils.easydict import EasyDict

    d = {'epochs': 100, 'len_dataloader': 58, 'name': 'four', 'save_path': 'save'}
    args = EasyDict(d)

    visualizer = Visualization(args)
    # visualizer.print_current_loss(1, 5, 0.088, 5.666)
    visualizer.plot_current_loss(2, 0.5, 1.8)
    #
    # dd = {'l': torch.rand((2, 3, 256, 256)),
    #       'r': torch.rand((2, 3, 256, 256)),
    #       'label': torch.rand((2, 3, 256, 256)),
    #       'y': torch.rand((2, 3, 256, 256))}
    #
    # visualizer.display_current_result(1,dd)



import numpy as np
from lbbutils.metrics.common import conv
from PIL import Image

np.seterr(divide='ignore', invalid='ignore')


def _q_g(im1: np.ndarray, im2: np.ndarray, fim: np.ndarray):
    """
        Ref: Objective Pixel-level Image Fusion Performance Measure, Proc. SPIE 4051, 89 (2000)
        by C. Xydeas
    :param im1:
    :param im2:
    :param fim:
    :return:
    """
    ker1 = np.array([[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]])
    ker2 = np.array([[-1, -2, -1], [0, 0, 0], [1, 2, 1]])
    # get the map
    fuseG, fuseA = get_map(fim, ker1, ker2)
    im1G, im1A = get_map(im1, ker1, ker2)
    im2G, im2A = get_map(im2, ker1, ker2)
    # edge preservation estimation
    Gaf, Aaf = edge_pres_est(fuseG, fuseA, im1G, im1A)
    Gbf, Abf = edge_pres_est(fuseG, fuseA, im2G, im2A)

    Qaf = f(Gaf, Aaf)
    Qbf = f(Gbf, Abf)

    Wa = im1G
    Wb = im2G
    tmp = (Qaf * Wa + Qbf * Wb) / (Wa + Wb)
    res=np.nanmean(tmp)
    return res


def f(G_f, A_f):
    gama1 = 1
    gama2 = 1
    k1 = -10
    k2 = -20
    delta1 = 0.5
    delta2 = 0.75
    Qg_AF = gama1 / (1 + np.exp(k1 * (G_f - delta1)))
    Qalpha_AF = gama2 / (1 + np.exp(k2 * (A_f - delta2)))
    Q_f = Qg_AF * Qalpha_AF
    return Q_f


def edge_pres_est(fuseG, fuseA, imG, imA):
    bimap = (imG > fuseG).astype(np.float64)
    buffer = (imG == 0).astype(np.float64)
    buffer = buffer * 0.00001
    imG = imG + buffer
    buffer1 = fuseG / imG
    buffer = (fuseG == 0).astype(np.float64)
    buffer = buffer * 0.00001
    fuseG = fuseG + buffer
    buffer2 = imG / fuseG
    Gaf = bimap * buffer1 + (1 - bimap) * buffer2
    Aaf = abs(abs(imA - fuseA) - np.pi / 2) * 2 / np.pi
    return Gaf, Aaf


def get_map(matrix, ker1, ker2):
    mat_X = conv(matrix, ker1, padding=1)
    mat_Y = conv(matrix, ker2, padding=1)
    mat_G = np.sqrt(mat_X * mat_X + mat_Y * mat_Y)
    buffer = (mat_X == 0).astype(np.float64)
    buffer = buffer * 0.00001
    mat_X = mat_X + buffer
    mat_A = np.arctan(mat_Y / mat_X)
    return mat_G, mat_A


if __name__ == '__main__':
    m1 = np.array(Image.open('../test/res/left.png'), dtype=np.float64)
    m2 = np.array(Image.open('../test/res/right.png'), dtype=np.float64)
    fim = np.array(Image.open('../test/res/fim.png'), dtype=np.float64)
    res = _q_g(m1, m2, fim)
    a = 1



import numpy as np


def normalize(x: np.ndarray):
    """

    :param x numpy.ndarray: x.shape (w,h,1)
    :return res numpy.ndarray: res.shape (w,h) astype=int16
    """
    if len(x.shape) == 2:
        data = x
        fz = data - np.min(data)
        fm = np.max(data) - np.min(data)
        res = fz / fm * 255
        return res.astype(np.int16)


def conv(im: np.ndarray, ker: np.ndarray, stride=1, padding=0):
    """

    :param im numpy.ndarray: (w,h) 2-D
    :param ker numpy.ndarray: (w,h) 2-D
    :param stride:
    :param padding: 1 is 'same',0 is 'valid'.
    :return:
    """
    r, c = im.shape
    pad_im = np.zeros((r + 2 * padding, c + 2 * padding))
    r_p, c_p = pad_im.shape
    pad_im[padding:r_p - padding, padding:c_p - padding] = im
    im = pad_im
    out_size = (r_p - ker.shape[0]) // stride + 1
    ret = np.zeros((out_size, out_size))
    for ri in range(0, out_size * stride, stride):
        for ci in range(0, out_size * stride, stride):
            region = im[ri:ri + ker.shape[0], ci:ci + ker.shape[0]]
            ret[ri // stride, ci // stride] = np.sum(region * ker)
    return ret


def create_window(win_size):
    def gaussian(win_size, sigma):
        gauss = np.array([np.exp(-(x - win_size // 2) ** 2 / float(2 * sigma ** 2)) for x in range(win_size)])
        return gauss / np.sum(gauss)

    _1D_window = np.expand_dims(gaussian(win_size, 1.5), 1)
    _2D_window = np.dot(_1D_window, _1D_window.T)
    return _2D_window


def ssim_yang(im1, im2):
    C1 = 2e-16;
    C2 = 2e-16;
    window = create_window(7)
    window = window / np.sum(window)
    mu1 = conv(im1, window)
    mu2 = conv(im2, window)
    mu1_sq = mu1 * mu1
    mu2_sq = mu2 * mu2
    mu1_mu2 = mu1 * mu2
    sigma1_sq = conv(im1 * im1, window) - mu1_sq
    sigma2_sq = conv(im2 * im2, window) - mu2_sq
    sigma12 = conv(im1 * im2, window) - mu1_mu2
    ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2))
    return ssim_map, sigma1_sq, sigma2_sq


def freqspace(num):
    lam = lambda x: np.arange(-x + 1, x, 2) / x if x // 2 == 1 else np.arange(-x, x - 1, 2) / x
    f1 = lam(num[0])
    f2 = lam(num[1])

    m, n = np.meshgrid(f1, f2)
    return n.T, m.T


if __name__ == '__main__':
    import torch

    # # (1)normalize
    # x = torch.rand((1, 1, 5, 5))
    # lam = lambda x: x.data.squeeze(0).permute((1, 2, 0)).numpy()
    # x_ = lam(x)
    # ret = normalize(x_)
    # a = 1

    # (2) conv
    # im = np.array([[1, 1, 1, 1, 1],
    #                [2, 2, 2, 2, 2],
    #                [3, 3, 3, 3, 3],
    #                [4, 3, 4, 3, 2],
    #                [3, 1, 5, 7, 3]])  # 1,1,1,1,1;2,2,2,2,2;3,3,3,3,3
    # ker = np.array([[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]])  # -1 0 1 ; -2 0 2 ; -1 0 1
    # ret = conv(im, ker, padding=0)

    # (3) create_window
    # create_window(7)

    # (4) ssim_yang
    # from PIL import Image
    #
    # m1 = np.array(Image.open('../test/fused1_ours.png'), dtype=np.float64)[:, :, 0]
    # m2 = np.array(Image.open('../test/fused2_ours.png'), dtype=np.float64)[:, :, 0]
    # fim = np.array(Image.open('../test/fused3_ours.png'), dtype=np.float64)[:, :, 0]
    # ssim_yang(m1, m2)

    # (5) freqspace()
    u, v = freqspace([3, 4])
    a = 1



import numpy as np


def _mi(im1, im2):
    im1_ = im1.astype(np.int32)
    im2_ = im2.astype(np.int32)
    row, col = im1.shape
    h = np.zeros((256, 256))
    for i in range(row):
        for j in range(col):
            h[im1_[i, j], im2_[i, j]] = h[im1_[i, j], im2_[i, j]] + 1
    # TODO 边缘直方图
    h = h / np.sum(h)
    im1_marg = np.sum(h, 0)  # 求列和
    im2_marg = np.sum(h, 1)  # 求行和

    lam = lambda x: (x == 0).astype(np.float64)

    H_x = -np.sum(im1_marg * np.log2(im1_marg + lam(im1_marg)))
    H_y = -np.sum(im2_marg * np.log2(im2_marg + lam(im2_marg)))
    H_xy = -np.sum(h * np.log2(h + lam(h)))
    mi = H_x + H_y - H_xy
    return mi, H_xy, H_x, H_y



import numpy as np

from .mi import _mi
from .q_cb import _q_cb as Qcb
from .q_g import _q_g as Qg
from .q_mi import _q_mi as Qim
from .q_y import _q_y as Qy


def metrics(l, r, pred):
    lamb = lambda x: np.array(x.squeeze(), dtype=np.float64)
    l, r, pred = lamb(l), lamb(r), lamb(pred)
    return Qim(l, r, pred), Qg(l, r, pred), Qy(l, r, pred), Qcb(l, r, pred)



from lbbutils.metrics.common import normalize, freqspace, conv
import numpy as np
from numpy.fft import fft2, fftshift, ifftshift, ifft2


def _q_cb(im1, im2, fim):
    """
        Ref: A new automated quality assessment algorithm for image fusion, Image and Vision Computing, 27 (2009) 1421-1432
        By Yin Chen et al.
    :param im1:
    :param im2:
    :param fim:
    :return:
    """
    im1 = im1 / 255
    im2 = im2 / 255
    fim = fim / 255

    im1 = normalize(im1)
    im2 = normalize(im2)
    fim = normalize(fim)

    f0 = 15.3870
    f1 = 1.3456
    a = 0.7622

    row, col = im1.shape
    HH, LL = row / 30, col / 30
    u, v = freqspace([row, col])
    u = LL * u
    v = HH * v
    r = np.sqrt(u * u + v * v)
    Sd = np.exp(-(r / f0) ** 2) - a * np.exp(-(r / f1) ** 2)

    fim1 = ifft2(ifftshift(fftshift(fft2(im1)) * Sd))
    fim2 = ifft2(ifftshift(fftshift(fft2(im2)) * Sd))
    ffim = ifft2(ifftshift(fftshift(fft2(fim)) * Sd))

    def gaussian2d(sigma):
        x, y = np.meshgrid(range(-15, 16), range(-15, 16))
        win = np.exp(-(x * x + y * y) / (2 * sigma * sigma)) / (2 * np.pi * sigma * sigma)
        return win

    def contrast(win1, win2, im):
        buff = conv(im, win1, padding=15)
        buff1 = conv(im, win2, padding=15)
        return buff / buff1 - 1

    G1 = gaussian2d(2)
    G2 = gaussian2d(4)

    k, p, q, h, Z = 1, 3, 2, 1, 0.0001
    C1 = contrast(G1, G2, np.real(fim1))
    C1 = abs(C1)
    C1P = (k * (np.power(C1, p))) / (h * (np.power(C1, q)) + Z)

    C2 = contrast(G1, G2, np.real(fim2))
    C2 = abs(C2)
    C2P = (k * (np.power(C2, p))) / (h * (np.power(C2, q)) + Z)

    Cf = contrast(G1, G2, np.real(ffim))
    Cf = abs(Cf)
    CfP = (k * (np.power(Cf, p))) / (h * (np.power(Cf, q)) + Z)

    mask = (C1P < CfP)
    mask = mask.astype(np.float64)
    Q1F = (C1P / CfP) * mask + (CfP / C1P) * (1 - mask)

    mask = (C2P < CfP)
    mask = mask.astype(np.float64)
    Q2F = (C2P / CfP) * mask + (CfP / C2P) * (1 - mask)

    ramda1 = (C1P * C1P) / (C1P * C1P + C2P * C2P)
    ramda2 = (C2P * C2P) / (C1P * C1P + C2P * C2P)
    Q = ramda1 * Q1F + ramda2 * Q2F
    ret = np.mean(Q)
    return ret


if __name__ == '__main__':
    from PIL import Image

    m1 = np.array(Image.open('../test/res/left.png'), dtype=np.float64)
    m2 = np.array(Image.open('../test/res/right.png'), dtype=np.float64)
    fim = np.array(Image.open('../test/res/fim.png'), dtype=np.float64)
    ret = _q_cb(m1, m2, fim)
    a = 1



import concurrent.futures as cf
import itertools
from fnmatch import fnmatchcase as match
from os import listdir
from os.path import join
from typing import Tuple, List, Any

import numpy as np
import torchvision.transforms.functional as F
from PIL import Image
from PIL import ImageFilter
from torch import Tensor
from torchvision.transforms import ToPILImage


class BlurImage(object):
    def __init__(self, num_div, style='C'):
        self.num_div = num_div
        self.style = style

    def __call__(self, image):
        l = Image.new('L', image.size)
        r = Image.new('L', image.size)
        blocks = []
        block_size = image.size[0] // self.num_div
        for row in range(1, self.num_div + 1):
            for col in range(1, self.num_div + 1):
                box = (block_size * (col - 1), block_size * (row - 1), block_size * (col - 1) + block_size + 1,
                       block_size * (row - 1) + block_size + 1)
                block = image.crop(box)
                blocks.append(block)
                blur_block = block.filter(ImageFilter.GaussianBlur(radius=3))
                if self.style == 'C':
                    if row % 2 == 1:
                        if (self.num_div * (row - 1) + (col - 1)) % 2 == 0:
                            l.paste(blur_block, box)
                            r.paste(block, box)
                        else:
                            l.paste(block, box)
                            r.paste(blur_block, box)
                    else:
                        if (self.num_div * (row - 1) + (col - 1)) % 2 == 1:
                            l.paste(blur_block, box)
                            r.paste(block, box)
                        else:
                            l.paste(block, box)
                            r.paste(blur_block, box)
                elif self.style == 'P':
                    if (self.num_div * (row - 1) + (col - 1)) % 2 == 0:
                        l.paste(blur_block, box)
                        r.paste(block, box)
                    else:
                        l.paste(block, box)
                        r.paste(blur_block, box)
        return l, r, image


class BlurImageRandomly:
    def __init__(self, radius=3, sigma=1):
        self.radius = radius
        self.sigma = sigma

    def __call__(self, image):
        im = np.array(image)
        left = im.copy()
        right = im.copy()
        row, col = im.shape
        filter = get_blur_filter(self.radius, self.sigma)
        mask = np.random.randint(0, 2, (row, col))

        for i in range(row):
            for j in range(col):
                if mask[j, i] == 1:
                    get_one_blurpoint(left, (j, i), filter)
                if mask[j, i] == 0:
                    get_one_blurpoint(right, (j, i), filter)

        l, r, image = Image.fromarray(left), Image.fromarray(right), Image.fromarray(im)
        return l, r, image


class ToTensors(object):
    def __call__(self, pic):
        return F.to_tensor(pic[0]), F.to_tensor(pic[1]), F.to_tensor(pic[2])


def save_img(imgs: Tuple[Tensor]) -> List[Any]:
    ret = list()
    for i, img in enumerate(imgs):
        if img.is_cuda:
            img = img.cpu().clone()
        else:
            img = img.clone()
        img.squeeze_(0)
        img_PIL = ToPILImage()(img)
        ret.append(img_PIL)
    return ret


def path_match(root, word):
    items = listdir(root)
    tmp = [item for item in items if match(item, '*-' + word + '-*')]
    return join(root, tmp[-1])


def gaussian2D(x, y, sigma=1):
    return 1 / (2 * np.pi * sigma ** 2) * np.exp(-(x ** 2 + y ** 2) / 2 * sigma ** 2)


def get_blur_filter(r, sigma=1):
    size = 2 * r + 1
    filter = np.zeros((size, size))
    for row in range(size):
        for col in range(size):
            filter[row, col] = gaussian2D(row - r, col - r, sigma=sigma)
    sum = np.sum(filter)
    filter = filter / sum
    return filter


def get_one_blurpoint(img, point, filter):
    r = filter.shape[0] // 2
    height, width = img.shape
    c_x, c_y = point
    min_x, min_y = c_x - r, c_y - r
    max_x, max_y = c_x + r, c_y + r
    tmp = 0
    for i in range(min_y, max_y + 1):
        for j in range(min_x, max_x + 1):
            now_x, now_y = j, i
            if j < 0:
                now_x = -j
            if i < 0:
                now_y = -i
            if j > width - 1:
                now_x = width - 1 - j
            if i > height - 1:
                now_y = height - 1 - i
            tmp += filter[i - min_y, j - min_x] * img[now_x, now_y]
    img[c_x, c_y] = tmp
    return img


if __name__ == '__main__':
    img = Image.open('test/res/fim.png')
    # im = np.array(img)
    # ret = np.zeros_like(im)
    # row, col = im.shape
    # filter = get_blur_filter(3)
    # for i in range(row):
    #     for j in range(col):
    #         get_one_blurpoint(im, (j, i), filter)
    # res = Image.fromarray(im)
    # res.save('test/res/fim_blured3_sigma1.png')
    bir = BlurImageRandomly()
    l, r, image = bir(img)
    l.save('test/res/bir_l.png')
    r.save('test/res/bir_r.png')
    image.save('test/res/bir_img.png')



"""异步调用

    开启一个子线程去执行该函数

    :parameter
        fnc (function): 执行方法
        callback (function): 回调方法
    :return
        pass
    Example:
        def multi(x):
            print(5 * x)

        @async_call(callback=multi)
        def add(a, b):
            time.sleep(5)
            return a + b

        def subtract(a, b):
            print(a - b)

        x=add(1,2)
        y=subtract(5, 3)

        Run:
            2
            15
"""
import threading
import time


class AsyncCall(object):
    def __init__(self, callable, callback=None):
        self.callable = callable
        self.callback = callback
        self.result = None

    def __call__(self, *args, **kwargs):
        thread = threading.Thread(target=self.run, args=args, kwargs=kwargs)
        thread.start()

    def run(self, *args, **kwargs):
        self.result = self.callable(*args, **kwargs)
        if self.callback:
            self.result = self.callback(self.result)


def async_call(callback):
    def deco(fnc):
        def __wrapper(*args, **kwargs):
            callback_ = AsyncCall(fnc, callback)(*args, **kwargs)
            return callback_

        return __wrapper

    return deco


if __name__ == '__main__':
    def multi(x):
        print(5 * x)


    @async_call(callback=multi)
    def add(a, b):
        time.sleep(5)
        return a + b


    def subtract(a, b):
        print(a - b)


    x = add(1, 2)
    y = subtract(5, 3)



from math import exp

import torch
import torch.nn.functional as F
from torch.autograd import Variable


class SSIM(torch.nn.Module):
    def __init__(self, window_size=11, size_average=True):
        super(SSIM, self).__init__()
        self.window_size = window_size
        self.size_average = size_average
        # self.channel = 1
        # self.window = create_window(window_size, self.channel)

    def forward(self, img1, img2):
        (_, channel, _, _) = img1.size()

        def create_window(window_size, channel):
            def gaussian(window_size, sigma):
                gauss = torch.Tensor([exp(-(x - window_size // 2) ** 2 / float(2 * sigma ** 2)) for x in range(window_size)])
                return gauss / gauss.sum()

            _1D_window = gaussian(window_size, 1.5).unsqueeze(1)
            _2D_window = _1D_window.mm(_1D_window.t()).float().unsqueeze(0).unsqueeze(0)
            window = Variable(_2D_window.expand(channel, 1, window_size, window_size).contiguous())
            return window

        window = create_window(self.window_size, channel)
        if img1.is_cuda:
            window = window.cuda(img1.get_device())
        window = window.type_as(img1)
        self.window = window
        self.channel = channel

        def _ssim(img1, img2, window, window_size, channel, size_average=True):
            mu1 = F.conv2d(img1, window, padding=window_size // 2, groups=channel)
            mu2 = F.conv2d(img2, window, padding=window_size // 2, groups=channel)

            mu1_sq = mu1.pow(2)
            mu2_sq = mu2.pow(2)
            mu1_mu2 = mu1 * mu2

            sigma1_sq = F.conv2d(img1 * img1, window, padding=window_size // 2, groups=channel) - mu1_sq
            sigma2_sq = F.conv2d(img2 * img2, window, padding=window_size // 2, groups=channel) - mu2_sq
            sigma12 = F.conv2d(img1 * img2, window, padding=window_size // 2, groups=channel) - mu1_mu2

            C1 = 0.01 ** 2
            C2 = 0.03 ** 2

            ssim_map = ((2 * mu1_mu2 + C1) * (2 * sigma12 + C2)) / ((mu1_sq + mu2_sq + C1) * (sigma1_sq + sigma2_sq + C2))

            if size_average:
                return ssim_map.mean()
            else:
                return ssim_map.mean(1).mean(1).mean(1)

        return _ssim(img1, img2, window, self.window_size, channel, self.size_average)


class TVLoss(torch.nn.Module):
    def __init__(self, tv_loss_weight=1):
        super(TVLoss, self).__init__()
        self.tv_loss_weight = tv_loss_weight

    def forward(self, x):
        batch_size = x.size()[0]
        h_x = x.size()[2]
        w_x = x.size()[3]
        count_h = self.tensor_size(x[:, :, 1:, :])
        count_w = self.tensor_size(x[:, :, :, 1:])
        h_tv = torch.pow((x[:, :, 1:, :] - x[:, :, :h_x - 1, :]), 2).sum()
        w_tv = torch.pow((x[:, :, :, 1:] - x[:, :, :, :w_x - 1]), 2).sum()
        return self.tv_loss_weight * 2 * (h_tv / count_h + w_tv / count_w) / batch_size

    @staticmethod
    def tensor_size(t):
        return t.size()[1] * t.size()[2] * t.size()[3]


class ColorLoss(torch.nn.Module):
    def __init__(self):
        super(ColorLoss, self).__init__()

    def forward(self, img1, img2):
        n, c, h, w = img1.shape
        vec1 = img1.view((n, -1, 3))
        vec2 = img2.view((n, -1, 3))
        norm_vec1 = vec1 / torch.norm(vec1, dim=2).unsqueeze(2).repeat(1, 1, 3)
        norm_vec2 = vec2 / torch.norm(vec2, dim=2).unsqueeze(2).repeat(1, 1, 3)
        tmp = torch.sum(torch.abs(norm_vec1 - norm_vec2), dim=2)
        tmp2 = torch.clamp(tmp, -0.999999, 0.999999)
        tmp3 = torch.acos(tmp2)
        return torch.mean(tmp3)


if __name__ == '__main__':
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    x1 = torch.rand((1, 3, 100, 100), requires_grad=True).to(device)
    x2 = torch.rand((1, 3, 100, 100)).to(device)
    # print(x1)
    loss = SSIM()
    res = loss(x1, x2)
    print(res)
    res.backward()
    # print(x1)



from codecs import open

from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

setup(
    name='lbbutils',

    version='0.0.16',

    description='lbbutils - Some tools for neural network and image fusion.',
    long_description=long_description,

    url='https://github.com/littlebaba/lbbutils',

    author='Heng',
    author_email='littlebaba0304@163.com',

    license='MIT',

    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],

    keywords='deep-learning multi-focus',

    packages=['lbbutils','lbbutils/metrics'],
    install_requires=[
        'numpy',
    ]
)



import time


# 装饰其第一种写法
# def timer(f):
#     def __wrapper(*args, **kwargs):
#         time_start = time.time()
#         result = f(*args, **kwargs)
#         time_end = time.time()
#         print(f'func:{f.__name__!r}  took: {time_end - time_start:{7}.{6}} sec')
#         return result
#
#     return __wrapper

# 装饰其第二种写法
class timer:
    """ 计算一个函数的执行时间
        Note:
            无
        Args:
            fnc (function): fucntions to be called.
        Example:
            ::
                @timer
                def foo(x):
                    for i in range(x):
                        time.sleep(0.01)
        """
    def __init__(self, fnc):
        self.fnc = fnc

    def __call__(self, *args, **kwargs):
        time_start = time.time()
        self.fnc(*args, **kwargs)
        time_end = time.time()
        print(f'func:{self.fnc.__name__!r}  took: {time_end - time_start:{7}.{6}} sec')


if __name__ == '__main__':

    @timer
    def foo(x):
        for i in range(x):
            time.sleep(0.01)


    foo(100)



from lbbutils.metrics.common import ssim_yang
import numpy as np


def _q_y(im1, im2, fim):
    """
        Ref: A novel similarity based quality metric for image fusion, Information Fusion, Vol.9, pp156-160, 2008
        by Cui Yang et al.
    :param im1:
    :param im2:
    :param fim:
    :return:
    """
    ssim_map1, sigma1_sq1, sigma2_sq1 = ssim_yang(im1, im2)
    ssim_map2, sigma1_sq2, sigma2_sq2 = ssim_yang(im1, fim)
    ssim_map3, sigma1_sq3, sigma2_sq3 = ssim_yang(im2, fim)
    boo_map = ssim_map1 >= 0.75
    ramda = sigma1_sq1 / (sigma1_sq1 + sigma2_sq1)
    Q1 = (ramda * ssim_map2 + (1 - ramda) * ssim_map3) * boo_map.astype(np.int32)
    Q2 = (np.maximum(ssim_map2, ssim_map3)) * ((~boo_map).astype(np.int32))
    Q = np.mean(Q1 + Q2)
    return Q


if __name__ == '__main__':
    from PIL import Image

    m1 = np.array(Image.open('../test/res/left.png'), dtype=np.float64)
    m2 = np.array(Image.open('../test/res/right.png'), dtype=np.float64)
    fim = np.array(Image.open('../test/res/fim.png'), dtype=np.float64)
    ret = _q_y(m1, m2, fim)
    a = 1



import time
import sys


class ProgresBar:
    def __init__(self, total_length):
        self.real_length = 40
        self.total_length = total_length
        self.c = self.real_length / total_length
        self.tmp = 0

    def increment(self):
        self.tmp += 1
        if self.tmp <= self.total_length:
            p = int(self.c * self.tmp)
            p_ = self.real_length - p
            str_bar = '\rProgress [' + '#' * p + '-' * p_ + ']'
            str_p = self.tmp / self.total_length
            str_all = str_bar + f'\t{str_p:{3}.{1}%}' + f'\t[{self.tmp}/{self.total_length}]'
            print(str_all, end='')


if __name__ == '__main__':
    bar = ProgresBar(145)
    for i in range(165):
        time.sleep(0.02)
        bar.increment()



import hashlib
import pickle
from typing import *
import os


class cache:
    """将一个函数的结果储存
        Note:
            pass
        Args:
            fnc (function): 被调用的函数
        Example:
                @cache
                def foo(x):
                    for i in range(10):
                        i *= (i + 1)
                    return i
    """

    def __init__(self, fnc: Callable):
        self.fnc = fnc
        self.fnc_name = self.fnc.__name__

    def __call__(self, *args, **kwargs):
        sha = hashlib.sha256()
        sha.update(pickle.dumps((self.fnc_name, args, frozenset(kwargs.items()))))
        out_path = os.path.join('.cache', f'{sha.hexdigest()}_{self.fnc_name}')
        try:
            with open(out_path, 'rb') as f:
                data = pickle.load(f)
        except FileNotFoundError:
            os.makedirs(os.path.dirname(out_path), exist_ok=True)
            data = self.fnc(*args, **kwargs)
            with open(out_path, 'wb') as f:
                pickle.dump(data, f)
        return data


if __name__ == '__main__':
    @cache
    def foo(x):
        for i in range(x):
            i *= (i + 1)
        return i


    print(foo(10))
    print(foo(101))
